@@ -9,15 +9,15 @@ Self = TypeVar("Self", bound="MutableSequenceIsliceProxy")
 
 
 class MutableSequenceIsliceProxy(SequenceIsliceProxy[T], Generic[T]):
-    _sequence: "src._viewable_mutable_sequence.ViewableMutableSequence[T]"
+    _sequence: "src.viewable_mutable_sequence.ViewableMutableSequence[T]"
 
     __slots__ = ()
 
-    def __init__(self: Self, sequence: "src._viewable_mutable_sequence.ViewableMutableSequence[T]", /) -> None:
-        assert isinstance(sequence, src._viewable_mutable_sequence.ViewableMutableSequence)
+    def __init__(self: Self, sequence: "src.viewable_mutable_sequence.ViewableMutableSequence[T]", /) -> None:
+        assert isinstance(sequence, src.viewable_mutable_sequence.ViewableMutableSequence)
         self._sequence = sequence
 
-    def __getitem__(self: Self, index: slice, /) -> "src._mutable_sequence_islice.MutableSequenceIslice[T]":
+    def __getitem__(self: Self, index: slice, /) -> "src.mutable_sequence_islice.MutableSequenceIslice[T]":
         if not isinstance(index, slice):
             raise TypeError(f"expected sequence.islice[start:stop:step], got sequence.islice[{index!r}]")
         # Cast integer-like indices to integers.
@@ -26,4 +26,4 @@ class MutableSequenceIsliceProxy(SequenceIsliceProxy[T], Generic[T]):
         start = None if index.start is None else range_.start
         stop = None if index.stop is None else range_.stop
         step = None if index.step is None else range_.step
-        return src._mutable_sequence_islice.MutableSequenceIslice(self._sequence, start, stop, step)
+        return src.mutable_sequence_islice.MutableSequenceIslice(self._sequence, start, stop, step)
